"""NapCat 事件解析

将 OB11 原始 JSON 转换为 BaseEventData 数据模型。
只产出纯数据模型，不创建实体。
"""

from __future__ import annotations

from typing import Any, Dict, Optional, Tuple, Type

from ncatbot.types import BaseEventData
from ncatbot.types.qq import (
    FriendRequestEventData,
    GroupMessageEventData,
    GroupRequestEventData,
    HeartbeatMetaEventData,
    LifecycleMetaEventData,
    MetaEventType,
    MessageType,
    NoticeType,
    NotifySubType,
    PostType,
    PrivateMessageEventData,
    RequestType,
    # notice
    FriendAddNoticeEventData,
    FriendRecallNoticeEventData,
    GroupAdminNoticeEventData,
    GroupBanNoticeEventData,
    GroupDecreaseNoticeEventData,
    GroupIncreaseNoticeEventData,
    GroupRecallNoticeEventData,
    GroupUploadNoticeEventData,
    HonorNotifyEventData,
    LuckyKingNotifyEventData,
    PokeNotifyEventData,
)
from ncatbot.utils import get_log

LOG = get_log("NapCatEventParser")


class EventParser:
    """事件解析器 — dict → types/ 数据模型

    使用 (post_type, secondary_key) 二元组管理注册表。
    """

    _registry: Dict[Tuple[str, str], Type[BaseEventData]] = {}

    @classmethod
    def register(cls, post_type: str, secondary_key: str = "") -> Any:
        def wrapper(data_cls: Type[BaseEventData]) -> Type[BaseEventData]:
            cls._registry[(post_type, secondary_key)] = data_cls
            return data_cls

        return wrapper

    @classmethod
    def _get_key(cls, data: Dict[str, Any]) -> Optional[Tuple[str, str]]:
        post_type = data.get("post_type", "")

        if post_type in (PostType.MESSAGE, "message_sent"):
            return (PostType.MESSAGE, data.get("message_type", ""))

        if post_type == PostType.NOTICE:
            notice_type = data.get("notice_type", "")
            if notice_type == NoticeType.NOTIFY:
                return (PostType.NOTICE, data.get("sub_type", ""))
            return (PostType.NOTICE, notice_type)

        if post_type == PostType.REQUEST:
            return (PostType.REQUEST, data.get("request_type", ""))

        if post_type == PostType.META_EVENT:
            return (PostType.META_EVENT, data.get("meta_event_type", ""))

        return None

    @classmethod
    def parse(cls, data: Dict[str, Any]) -> BaseEventData:
        """解析 raw dict → 数据模型实例"""
        key = cls._get_key(data)
        if not key:
            raise ValueError(f"Unknown event: post_type={data.get('post_type')}")
        data_cls = cls._registry.get(key)
        if not data_cls:
            raise ValueError(f"No data class registered for {key}")
        return data_cls.model_validate(data)


def _register_builtin() -> None:
    """注册内置事件数据模型"""
    r = EventParser.register

    # Message
    r(PostType.MESSAGE, MessageType.PRIVATE)(PrivateMessageEventData)
    r(PostType.MESSAGE, MessageType.GROUP)(GroupMessageEventData)

    # Request
    r(PostType.REQUEST, RequestType.FRIEND)(FriendRequestEventData)
    r(PostType.REQUEST, RequestType.GROUP)(GroupRequestEventData)

    # Meta
    r(PostType.META_EVENT, MetaEventType.LIFECYCLE)(LifecycleMetaEventData)
    r(PostType.META_EVENT, MetaEventType.HEARTBEAT)(HeartbeatMetaEventData)

    # Notice
    r(PostType.NOTICE, NoticeType.GROUP_UPLOAD)(GroupUploadNoticeEventData)
    r(PostType.NOTICE, NoticeType.GROUP_ADMIN)(GroupAdminNoticeEventData)
    r(PostType.NOTICE, NoticeType.GROUP_DECREASE)(GroupDecreaseNoticeEventData)
    r(PostType.NOTICE, NoticeType.GROUP_INCREASE)(GroupIncreaseNoticeEventData)
    r(PostType.NOTICE, NoticeType.GROUP_BAN)(GroupBanNoticeEventData)
    r(PostType.NOTICE, NoticeType.FRIEND_ADD)(FriendAddNoticeEventData)
    r(PostType.NOTICE, NoticeType.GROUP_RECALL)(GroupRecallNoticeEventData)
    r(PostType.NOTICE, NoticeType.FRIEND_RECALL)(FriendRecallNoticeEventData)

    # Notice - Notify
    r(PostType.NOTICE, NotifySubType.POKE)(PokeNotifyEventData)
    r(PostType.NOTICE, NotifySubType.LUCKY_KING)(LuckyKingNotifyEventData)
    r(PostType.NOTICE, NotifySubType.HONOR)(HonorNotifyEventData)


_register_builtin()


class NapCatEventParser:
    """将 OB11 原始 JSON 转换为 BaseEventData"""

    def parse(self, raw_data: dict) -> Optional[BaseEventData]:
        if "post_type" not in raw_data:
            return None
        raw_data.setdefault("platform", "qq")
        try:
            return EventParser.parse(raw_data)
        except ValueError as e:
            LOG.debug(f"事件解析跳过: {e}")
            return None
        except Exception as e:
            LOG.warning(f"事件解析失败: {e}")
            return None



"""直播间数据源 — 每个实例在独立线程中监听一个直播间的 WebSocket 弹幕流

每个 LiveSource 拥有独立的线程和事件循环，避免多房间或
LiveDanmaku 内部阻塞操作影响主事件循环。事件通过
``asyncio.run_coroutine_threadsafe`` 线程安全地回调到主循环。
"""

from __future__ import annotations

import asyncio
import threading
from typing import Any, Awaitable, Callable, Optional

from ncatbot.utils import get_log

from .base import BaseSource

LOG = get_log("LiveSource")


class LiveSource(BaseSource):
    source_type = "live"

    def __init__(
        self,
        room_id: int,
        credential: "Any",
        callback: Callable[[str, dict], Awaitable[None]],
        *,
        max_retry: int = 5,
        retry_after: float = 1.0,
    ) -> None:
        super().__init__(callback)
        self.source_id = str(room_id)
        self._room_id = room_id
        self._credential = credential
        self._max_retry = max_retry
        self._retry_after = retry_after
        self._danmaku: Optional["Any"] = None
        self._thread: Optional[threading.Thread] = None
        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self._main_loop: Optional[asyncio.AbstractEventLoop] = None
        self._ready = threading.Event()

    async def start(self) -> None:
        if self._running:
            return

        self._main_loop = asyncio.get_running_loop()
        self._ready.clear()

        from bilibili_api.live import LiveDanmaku

        self._danmaku = LiveDanmaku(
            room_display_id=self._room_id,
            credential=self._credential,
            max_retry=self._max_retry,
            retry_after=self._retry_after,
        )

        @self._danmaku.on("ALL")
        async def _on_all(event: dict) -> None:
            # 开播事件：获取直播间信息附加到事件数据
            if event.get("type") == "LIVE":
                data = event.get("data", {})
                if isinstance(data, dict) and data.get("live_time", 0):
                    await self._attach_room_info(event)

            # 线程安全地将事件回调到主事件循环
            if self._main_loop is not None and self._main_loop.is_running():
                asyncio.run_coroutine_threadsafe(
                    self._callback("live", event), self._main_loop
                )

        self._running = True
        self._thread = threading.Thread(
            target=self._run_thread,
            name=f"LiveSource({self._room_id})",
            daemon=True,
        )
        self._thread.start()
        LOG.info("直播源 %s 已启动 (线程: %s)", self._room_id, self._thread.name)

    def _run_thread(self) -> None:
        """线程入口：创建独立事件循环并运行 LiveDanmaku 连接。"""
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        self._loop = loop
        try:
            loop.run_until_complete(self._connect_danmaku())
        except Exception:
            LOG.exception("直播源 %s 线程异常退出", self._room_id)
        finally:
            self._running = False
            try:
                loop.run_until_complete(loop.shutdown_asyncgens())
            except Exception:
                pass
            loop.close()
            self._loop = None

    async def _attach_room_info(self, event: dict) -> None:
        """在子线程事件循环中获取直播间信息并附加到事件数据。"""
        try:
            from bilibili_api.live import LiveRoom

            room = LiveRoom(
                room_display_id=self._room_id,
                credential=self._credential,
            )
            raw_info = await room.get_room_info()
            event["room_info"] = raw_info
            LOG.info(
                "直播源 %s 开播，已获取直播间信息: %s",
                self._room_id,
                raw_info.get("room_info", {}).get("title", ""),
            )
        except Exception:
            LOG.warning("直播源 %s 获取直播间信息失败", self._room_id, exc_info=True)

    async def _connect_danmaku(self) -> None:
        """在独立事件循环中运行弹幕连接。"""

        @self._danmaku.on("VERIFICATION_SUCCESSFUL")
        async def _on_connected(_: dict) -> None:
            self._ready.set()
            LOG.debug("直播源 %s WebSocket 验证成功", self._room_id)

        try:
            await self._danmaku.connect()
        except asyncio.CancelledError:
            pass
        except Exception:
            LOG.exception("直播源 %s 连接异常", self._room_id)
        finally:
            self._ready.set()  # 防止异常时 ready 永远阻塞

    async def stop(self) -> None:
        if self._danmaku is not None and self._loop is not None:
            loop = self._loop
            try:
                future = asyncio.run_coroutine_threadsafe(
                    self._danmaku.disconnect(), loop
                )
                future.result(timeout=10)
            except Exception:
                LOG.debug("断开直播源 %s 时异常", self._room_id, exc_info=True)
            # 通知子线程事件循环停止
            loop.call_soon_threadsafe(loop.stop)

        if self._thread is not None and self._thread.is_alive():
            self._thread.join(timeout=15)
            if self._thread.is_alive():
                LOG.warning("直播源 %s 线程未能在超时内退出", self._room_id)

        self._running = False
        self._danmaku = None
        self._thread = None
        self._loop = None
        LOG.info("直播源 %s 已停止", self._room_id)



"""
RBAC 实体管理模块 - 处理权限、角色、用户的管理
"""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .service import RBACService


class EntityManager:
    """权限、角色、用户管理器"""

    def __init__(self, service: "RBACService"):
        self._service = service

    # === 权限管理 ===

    def add_permission(self, path: str) -> None:
        if not self._service._permissions.exists(path, exact=True):
            self._service._permissions.add(path)
            self._service._clear_cache()

    def remove_permission(self, path: str) -> None:
        self._service._permissions.remove(path)
        self._service._clear_cache()

    def permission_exists(self, path: str) -> bool:
        return self._service._permissions.exists(path, exact=True)

    # === 角色管理 ===

    def add_role(self, role: str, exist_ok: bool = False) -> None:
        if role in self._service._roles:
            if not exist_ok:
                raise ValueError(f"角色 {role} 已存在")
            return

        self._service._roles[role] = {"whitelist": set(), "blacklist": set()}
        self._service._role_users[role] = set()
        self._service._role_inheritance[role] = []

    def remove_role(self, role: str) -> None:
        if role not in self._service._roles:
            raise ValueError(f"角色 {role} 不存在")

        for user_data in self._service._users.values():
            if role in user_data["roles"]:
                user_data["roles"].remove(role)

        for parent_roles in self._service._role_inheritance.values():
            if role in parent_roles:
                parent_roles.remove(role)

        del self._service._roles[role]
        del self._service._role_users[role]
        del self._service._role_inheritance[role]

    def role_exists(self, role: str) -> bool:
        return role in self._service._roles

    def set_role_inheritance(self, role: str, parent: str) -> None:
        if role not in self._service._roles:
            raise ValueError(f"角色 {role} 不存在")
        if parent not in self._service._roles:
            raise ValueError(f"父角色 {parent} 不存在")
        if role == parent:
            raise ValueError("不能继承自身")
        if self._would_create_cycle(role, parent):
            raise ValueError(f"检测到循环继承: {role} -> {parent}")

        if parent not in self._service._role_inheritance[role]:
            self._service._role_inheritance[role].append(parent)
            self._service._clear_cache()

    def _would_create_cycle(self, role: str, new_parent: str) -> bool:
        visited = set()

        def check(current: str) -> bool:
            if current == role:
                return True
            if current in visited:
                return False
            visited.add(current)
            for parent in self._service._role_inheritance.get(current, []):
                if check(parent):
                    return True
            return False

        return check(new_parent)

    # === 用户管理 ===

    def add_user(self, user: str, exist_ok: bool = False) -> None:
        if user in self._service._users:
            if not exist_ok:
                raise ValueError(f"用户 {user} 已存在")
            return

        roles = [self._service._default_role] if self._service._default_role else []
        self._service._users[user] = {
            "whitelist": set(),
            "blacklist": set(),
            "roles": roles,
        }

        if (
            self._service._default_role
            and self._service._default_role in self._service._role_users
        ):
            self._service._role_users[self._service._default_role].add(user)

    def remove_user(self, user: str) -> None:
        if user not in self._service._users:
            raise ValueError(f"用户 {user} 不存在")

        for role, users in self._service._role_users.items():
            users.discard(user)

        del self._service._users[user]
        self._service._clear_cache()

    def user_exists(self, user: str) -> bool:
        return user in self._service._users

    def user_has_role(self, user: str, role: str, create_user: bool = True) -> bool:
        if not self.user_exists(user):
            if create_user:
                self.add_user(user)
            else:
                return False

        all_roles = set()

        def collect_roles(r: str):
            if r in all_roles:
                return
            all_roles.add(r)
            for parent in self._service._role_inheritance.get(r, []):
                collect_roles(parent)

        for r in self._service._users[user]["roles"]:
            collect_roles(r)

        return role in all_roles

    def assign_role(
        self,
        user: str,
        role: str,
        create_user: bool = True,
    ) -> None:
        if role not in self._service._roles:
            raise ValueError(f"角色 {role} 不存在")

        if user not in self._service._users:
            if create_user:
                self.add_user(user)
            else:
                raise ValueError(f"用户 {user} 不存在")

        user_data = self._service._users[user]
        if role not in user_data["roles"]:
            user_data["roles"].append(role)

        if user not in self._service._role_users[role]:
            self._service._role_users[role].add(user)

        self._service._clear_cache()

    def unassign_role(self, user: str, role: str) -> None:
        if user not in self._service._users:
            raise ValueError(f"用户 {user} 不存在")

        if role not in self._service._roles:
            raise ValueError(f"角色 {role} 不存在")

        user_data = self._service._users[user]
        if role in user_data["roles"]:
            user_data["roles"].remove(role)

        self._service._role_users[role].discard(user)
        self._service._clear_cache()
